'' =================================================================================================
''
''   File....... jm_sbus_rx.spin
''   Purpose.... Futaba S.BUS receiver / decoder
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2021 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 23 MAY 2021
''
''   {$P1}
''
'' =================================================================================================

{
   S.BUS is a 25-byte packet sent at 100K baud, using inverted 8E2 format (9 bits)
   -- mark level is 0v
   -- space level is 3.3v

   [header] [data1] ... [data22] [flags] [footer]

   -- header is $0F and footer is $00

   Flags (byte 23)
   -- bit0 = Ch17
   -- bit1 = Ch18
   -- bit2 = frame lost
   -- bit3 = failsafe activated

   If a parity error is detected the frame lost frame flag will be set and the receiver/decoder
   will wait for the next packet.

   Special thanks to Tony B and Evan H in the forums for helpful suggestions that simplified
   the unpacking code and made it faster. Also to Jason Dorie for his S.BUS code that inspired
   some aspects of how I approached S.BUS reception and decoding.

   This version is ported from my P2 version, jm_sbus_rx.spin2. They are functionally equal.

   Tested with FlySky FS-iA10B receiver configured for S.BUS output (fast, 130 packets/second)

   23 MAY 2021: Emperical testing shows that at 80MHz, the busy flag will be set for ~24 microseconds
                while the background updates the hub positions array.
}


con { fixed io pins }

  PGM_RX = 31  { I }                                            ' serial / programming
  PGM_TX = 30  { O }

  EE_SDA = 29  { I/O }                                          ' i2c / eeprom
  EE_SCL = 28  { I/O }


con

  #true,  ON, OFF
  #false, NO, YES

   POS_MIN = $0000
   POS_MAX = $07FF

   POS_CTR = (POS_MAX - POS_MIN) >> 1


var

  long  cog                                                     ' cog running S.BUS receive code

  ' do not change order of these variables

  long  rxp         { par[0] }                                  ' from S.BUS pin on RC receiver
  long  baudticks   { par[1] }                                  ' ticks in baud
  long  us1ticks    { par[2] }                                  ' ticks in 1us
  long  p_chans     { par[3] }                                  ' pointer to channel data
  long  frameloss   { par[4] }                                  ' frame lost
  long  failsafe    { par[5] }                                  ' packet indicated failsafe mode
  long  packets     { par[6] }                                  ' # of packets written
  long  updating    { par[7] }                                  ' cog is writing to hub

  word  channel[18]                                             ' latest channel values


pub null

'' This is not a top level object


pub start(rxpin, p_defaults)

'' Start S.BUS receiver cog
'' -- rxpin is input from S.BUS receiver
'' -- p_defaults is pointer to table (words) of initial channel values
''    * optional (use -1 to ignore)

  stop                                                          ' stop if already running

  rxp       := rxpin                                            ' configure parameters for cog
  baudticks := clkfreq /   100_000                              ' ticks in 1 rx bit
  us1ticks  := clkfreq / 1_000_000                              ' ticks in 1us
  p_chans   := @channel                                         ' pointer to channels array

  if (p_defaults > 0)                                           ' start-up values specified?
    wordmove(@channel, p_defaults, 18)                          '  copy default values

  cog := cognew(@entry, @rxp) + 1                               ' start S.BUS rx cog

  return cog


pub stop

'' Stops S.BUS cog if running

  if (cog)                                                      ' if running
    cogstop(cog-1)                                              '  stop the cog
    cog := 0                                                    '  mark stopped

  longfill(@rxp, 0, 8)                                          ' clear object globals

  frameloss := true                                             ' set until good signal acquired

  wordfill(@channel, POS_CTR, 18)                               ' initialize positions


pub read(ch) : position

'' Returns channel value
'' -- channels are 1..18

  if ((ch < 1) or (ch > 18))
    return 0
  else
    return channel[ch-1]


pub busy

'' Returns true when cog is updating hub

  return updating


pub has_signal

'' Returns true when signal present

  return (frameloss == false)


pub in_failsafe

'' Returns true when in failsafe mode

  return failsafe


pub count

  return packets


pub address

'' Returns hub address of channel data

  return @channel


dat { s.bus receiver }

                org       0

entry           mov       t1, par                               ' hub address of parameters
                rdlong    t2, t1                                ' get input pin
                mov       rxmask, #1                            ' convert to mask
                shl       rxmask, t2
                andn      dira, rxmask                          ' force to input

                add       t1, #4
                rdlong    rxbit1x0, t1                          ' get rx 1.0 bit timing

                mov       rxbit1x5, rxbit1x0                    ' set rx 1.5 bit timing
                shr       rxbit1x5, #1
                add       rxbit1x5, rxbit1x0

                add       t1, #4
                rdlong    us1, t1                               ' get ticks in 1us

                add       t1, #4
                rdlong    p_ch, t1                              ' get channel array location

                add       t1, #4                                ' setup other hub pointers
                mov       p_los, t1

                add       t1, #4
                mov       p_failsafe, t1

                add       t1, #4
                mov       p_pktcnt, t1

                add       t1, #4
                mov       p_busy, t1


' top of main loop
' -- wait for rx pin to be quiet (low) for 250us (>2 byte periods)

wait_quiet      mov       t1, #250
                mov       t2, cnt
                add       t2, us1
:check_quiet    waitcnt   t2, us1
                test      rxmask, ina                   wc
    if_c        jmp       #wait_quiet
                djnz      t1, #:check_quiet


' receive, validate, and save s.bus bytes

                mov       sbusidx, #0

rx_byte         waitpeq   rxmask, rxmask                        ' wait for start bit
                mov       rxtimer, cnt                          ' sync timer
                add       rxtimer, rxbit1x5                     ' set timer to skip start bit

                mov       rxwork, #0                            ' clear rx workspace
                mov       rxbits, #9                            ' rx 9 bits (8 + parity)

get_bit         waitcnt   rxtimer, rxbit1x0                     ' wait for middle of bit
                test      rxmask, ina                   wc      ' sample rx
                shr       rxwork, #1                            ' prep for bit
                muxnc     rxwork, #%1_00000000                  ' !c -> rxwork.8
                djnz      rxbits, #get_bit                      ' more?

                waitpne   rxmask, rxmask                        ' wait for stop bit

                and       rxwork, #$1FF                 wc      ' check parity
    if_c        jmp       #bad_packet                           ' abort on bad byte

                and       rxwork, #$0FF                         ' strip parity bit

save_sbus       mov       sbusbyte, rxwork
                call      #wr_sbus
                add       sbusidx, #1
                cmpsub    sbusidx, #25                  wc      ' done
    if_nc       jmp       #rx_byte


' verify header ($0F) and footer ($00) bytes

verify          mov       sbusidx, #24                          ' footer
                call      #rd_sbus
                mov       t2, sbusbyte
                shl       t2, #8

                mov       sbusidx, #0                           ' header
                call      #rd_sbus
                or        t2, sbusbyte
                cmp       t2, #$000F                    wc, wz
    if_ne       jmp       #bad_packet                           ' if no match, wait for next packet


' unpack s.bus bytes to channel values

                wrlong    IS_ON, p_busy                         ' set updating flag

unpack          mov       p_hub, p_ch                           ' point to channels array
                mov       chcnt, #16                            ' set channels count
                mov       chwork, #0                            ' channel data workspace
                mov       chbc, #0                              ' bit count in workspace
                mov       sbusidx, #1                           ' s.bus data index (cog)

build_ch        call      #rd_sbus                              ' get byte from cog array
                add       sbusidx, #1                           ' bump cog index
                shl       sbusbyte, chbc                        ' prep new byte
                or        chwork, sbusbyte                      ' add to channel value
                add       chbc, #8                              ' update bit count
                cmpsub    chbc, #11                     wc      ' enough bits for channel value?
    if_nc       jmp       #build_ch

                mov       t1, chwork                            ' make copy
                and       t1, CH_MAX                            ' strip extra bits
                wrword    t1, p_hub                             ' write to hub
                add       p_hub, #2                             ' update hub pointer
                shr       chwork, #11                           ' remove last channel
                djnz      chcnt, #build_ch                      ' done?


' get digital channels from flags byte

                mov       sbusidx, #23
                call      #rd_sbus                              ' get flags byte from cog array

                test      sbusbyte, #%0000_0001         wc      ' ch17 is digital
    if_nc       mov       chwork, #0                            ' if nc, make 0
    if_c        mov       chwork, CH_MAX                        ' of c, make 2047
                wrword    chwork, p_hub                         ' write to hub
                add       p_hub, #2                             ' bump hub pointer

                test      sbusbyte, #%0000_0010         wc      ' ch18 is digital
    if_nc       mov       chwork, #0                            ' if nc, make 0
    if_c        mov       chwork, CH_MAX                        ' of c, make 2047
                wrword    chwork, p_hub                         ' write to hub


' check loss-of-signal and failsafe flags

chk_los         test      sbusbyte, #%0000_0100         wc
    if_nc       mov       t1, #0                                ' if nc, make false
    if_c        neg       t1, #1                                ' if c, make true
                wrlong    t1, p_los                             ' write to hub

chk_failsafe    test      sbusbyte, #%0000_1000         wc
    if_nc       mov       t1, #0                                ' if nc, make false
    if_c        neg       t1, #1                                ' if c, make true
                wrlong    t1, p_failsafe                        ' write to hub

                rdlong    t1, p_pktcnt                          ' update packet count
                add       t1, #1
                wrlong    t1, p_pktcnt

                wrlong    IS_OFF, p_busy                        ' clear updating flag

                jmp       #wait_quiet                           ' back to top


' if problem with packet, set lostframe flag

bad_packet      wrlong    IS_ON, p_los                          ' set lost frame flag
                jmp       #wait_quiet                           ' wait for next packet



' write value in sbusbyte to sbusraw[sbusidx]

wr_sbus         mov       t3, #sbusraw                          ' point to sbus array
                add       t3, sbusidx                           ' add index
                movd      :write, t3                            ' update :write instruction
                nop
:write          mov       0-0, sbusbyte                         ' write value to table
wr_sbus_ret     ret


' read value from sbusraw[sbusidx] to sbusbyte

rd_sbus         mov       t3, #sbusraw                          ' point to sbus array
                add       t3, sbusidx                           ' add index
                movs      :read, t3                             ' update :read instruction
                nop
:read           mov       sbusbyte, 0-0                         ' read value from table
rd_sbus_ret     ret


' -------------------------------------------------------------------------------------------------

IS_ON           long      -1
IS_OFF          long      0

CH_MAX          long      $07FF                                 ' mask for maximum value


p_ch            res       1                                     ' -> channel array
p_los           res       1                                     ' -> frame loss flag
p_failsafe      res       1                                     ' -> in failsafe mode
p_pktcnt        res       1                                     ' -> packet count
p_busy          res       1                                     ' -> updating

rxmask          res       1                                     ' rx pin mask
rxwork          res       1                                     ' workspace for incoming byte
rxbits          res       1                                     ' bits to receive
rxtimer         res       1                                     ' bit timer
rxbit1x0        res       1                                     ' ticks in bit
rxbit1x5        res       1                                     ' ticks in 1.5 bits

us1             res       1                                     ' ticks in 1us

sbuscnt         res       1                                     ' byte counter
sbusraw         res       25                                    ' raw sbus data (compressed)
sbusidx         res       1
sbusbyte        res       1


p_hub           res       1                                     ' pointer to channel
chcnt           res       1                                     ' channels count
chwork          res       1                                     ' channel value workspace
chbc            res       1                                     ' bit count in channel workspace

t1              res       1                                     ' temp vars
t2              res       1
t3              res       1

                fit       496


con { license }

{{

  MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}